local Scene = Object:extend()
require("monsterstate")

local WalkGraph = Object:extend()
require("jimmy")

function WalkGraph:findVertex(v)
  for j, v2 in ipairs(self.vertices) do
    if v2[1] == v[1] and v2[2] == v[2] then
      return v2
    end
  end
end

function WalkGraph:insertVertex(v)
  local v2 = self:findVertex(v)
  if not v2 then
    local new = {v[1], v[2], {}}
    self.vertices[#self.vertices+1] = new
    return new
  else
    return v2
  end
end


function WalkGraph:new(edges)
  self.vertices = {}
  self.edges = {}
  self.lasmatch = {0,0}
  self.lastclick = {0,0}
  
  for i, e in ipairs(edges) do
    local a = self:insertVertex(e[1])
    local b = self:insertVertex(e[2])
    local ne = {a,b,e[3]}
    self.edges[#self.edges+1] = ne
    a[3][#a[3]+1] = ne
    b[3][#b[3]+1] = ne
  end
end


function WalkGraph:findNearestPoint(x, y)
  local min = math.huge
  local minP
  local minE
  local sq = lust.tools.sq
  for i, e in ipairs(self.edges) do
    local px, py = lust.tools.point_line_pivot(x,y, e[1][1], e[1][2], e[2][1], e[2][2])
    local d = sq(px-x) + sq(py-y)
    if d < min then
      minP = {px, py}
      minE = e
      min = d
    end
  end

  return minP, minE
end


function WalkGraph:findNodePath(v1, v2)
  local unvisited = {}
  local current = v1
  local dest = v2
  for i, v in ipairs(self.vertices) do
    v[4] = math.huge
    v[5] = false
    unvisited[v] = v
  end

  current[4] = 0

  while not dest[5] do
    for i, e in ipairs(current[3]) do
      local f
      if e[1] == current then
        f = e[2]
      else
        f = e[1]
      end

      if f[4] > current[4] + 1 then
        f[6] = current
        f[7] = e
        f[4] = current[4] + 1
      end
    end
    current[5] = true
    unvisited[current] = nil

    local mind = math.huge
    for v in pairs(unvisited) do
      if v[4] < mind then
        mind = v[4]
        current = v
      end
    end
  end

  local path = {}
  while dest ~= v1 do
    table.insert(path, 1, {dest[1], dest[2], dest[7]})
    dest = dest[6]
  end

  table.insert(path, 1, {v1[1], v2[2]})

  return path
end


function WalkGraph:findPath(sx, sy, dx, dy)
  local ps, es = self:findNearestPoint(sx, sy)
  local pd, ed = self:findNearestPoint(dx, dy)

  local paths = {
    self:findNodePath(es[1], ed[1]),
    self:findNodePath(es[2], ed[1]),
    self:findNodePath(es[1], ed[2]),
    self:findNodePath(es[2], ed[2]),
  }

  table.sort(paths, function(a,b) return #a > #b end)
  local path = paths[1]

--[[
  if #path == 1 then
    return {{ps[1], ps[2]}, {pd[1], pd[2]}}
  end
  ]]

  path[1][1] = ps[1]
  path[1][2] = ps[2]
  path[#path][1] = pd[1]
  path[#path][2] = pd[2]

--[[
  -- fix beginning and end of path so they are the actual closest points on the graph
  if path[2][1] == es[2][1] and path[2][2] == es[2][2] then
    path[1][1] = ps[1]
    path[1][2] = ps[2]
  else
    table.insert(path, 1, {ps[1], ps[2]})
  end

  local nm1 = #path-1
  if path[nm1][1] == ed[2][1] and path[nm1][2] == ed[2][2] then
    path[nm1+1][1] = pd[1]
    path[nm1+1][2] = pd[2]
  else
    table.insert(path, nm1+2, {pd[1], pd[2]})
  end
  ]]

  return path
end


function WalkGraph:findValidPath(sx, sy, dx, dy)
  local path = self:findPath(sx, sy, dx, dy)

  local v = 0
  for i, p in ipairs(path) do
    if p[3]() then
      v = v + 1
    else
      break
    end
  end

  for i = #path,v, -1 do
    path[#path] = nil
  end

  return path
end


local factories = {
  Person = function(mngr, d) return Person(mngr, d.sprite, d.pos.x, d.pos.y) end,
  Jimmy = function(mngr, d)
    local j = Jimmy(mngr, d.pos.x, d.pos.y) 
    return j
  end
}


function Scene:new(state, chapter, player)
  self.shutter = 0.0
  self.state = state
  local def = require('chapters/' .. chapter)
  self.layers = {}
  for i,l in ipairs(def.layers) do
    if l == '<player>' then
      self.playerLayer = i
      self.layers[i] = "dummy"
    else
      self.layers[i] = love.graphics.newImage(l)
    end
  end

  self.layerDisabled = {}

  self.flux = flux:group()

  self.timekeeper = lust.TimeKeeper()


  self.walkgraph = WalkGraph(def.edges)
  self.edges = def.edges

  self.entities = lust.Group(self)
  self.player = Player(self, def.start.x, def.start.y)
  self.entities:add(self.player)
  self.timescale = 1.0

  self.names = {}
  for n, e in pairs(def.entities) do
    local ne = factories[e.type](self, e)
    self.entities:add(ne)
    self.names[n] = ne
  end

  self.dialogsystems = def.dialogs(self)
  self.triggers = def.triggers
  self.walkzones = def.walkzones
  self.clickzones = def.clickzones

  self.images = {}
  self.data = {}
  self.title = def.title
  self.triggers.start(self)

end


function Scene:startDialog(name)
  self.activeDialog = self.dialogsystems[name]
  self.activeDialog:open(self.player.data)
  love.mouse.setCursor(G.cursors.pointer)
end


function Scene:update(dt)
  self.entities:update(dt * self.timescale)
  if self.activeDialog and self.activeDialog:isOpen() then
    self.activeDialog:update(dt)
  else
    for _, z in pairs(self.walkzones) do
      if self.player.pos.x >= z[1] and self.player.pos.x <= z[3]
        and self.player.pos.y >= z[2] and self.player.pos.y <= z[4] then
        self.triggers[z[5]](self)
      end
    end
  end
  self.timekeeper:update(dt)
  self.flux:update(dt)
end


function Scene:removeWalkzone(name)
  self.walkzones[name] = nil
end

function Scene:removeClickzone(name)
  self.clickzones[name] = nil
end

function Scene:addClickzone(name, zone)
  self.clickzones[name] = zone
end


function Scene:addEntity(name, ent)
  self.entities:add(ent)
  self.names[name] = ent

end


function Scene:addImage(name, img, x, y, layer)
  local l = self.images[layer]
  if not l then
    l = {}
    self.images[layer] = l
  end

  local i = G.core.images:load(img)
  i:setFilter('nearest', 'nearest')
  l[name]= {i, x, y}
end


function Scene:start(data)
  data.pos.x = self.start.x
  data.pos.y = self.start.y
end


function Scene:click(x,y)
  if self.activeDialog and self.activeDialog:isOpen() then
    self.activeDialog:click(x,y)
  elseif self.haveOver then
    if self.overFlux then
      self.overFlux:stop()
      self.overPos = 1
      self.overFlux = nil
    else
      self.timekeeper:skip()
      self.haveOver = false
    end

  else
    local z = self.zone
    if not self.noWalking then
      if z then
        local p = self.walkgraph:findPath(self.player.pos.x, self.player.pos.y, z[8], z[9])
       -- table.remove(p, 1)
        self.player.path = p
        self.player.action = function() self.triggers[z[7]](self) end
      else
        local p = self.walkgraph:findPath(self.player.pos.x, self.player.pos.y, x,y-30)
        --table.remove(p, 1)
        self.player.path = p
        self.player.action = nil
      end
      if self.player.path and #self.player.path > 0 then
        if self.triggers.move then
          self.triggers.move(self)
        end
      end
    end
  end
end


function Scene:mousemoved(x, y)
  if self.activeDialog and self.activeDialog:isOpen() then
    self.activeDialog:mouseover(x,y)
  else
    local cursor = "pointer"
    y = y - 30
    self.zone = nil
    for _, z in pairs(self.clickzones) do
      if x >= z[1] and x <= z[3]
        and y >= z[2] and y <= z[4] then
        cursor = z[5]
        self.zone = z
      end
    end
    love.mouse.setCursor(G.cursors[cursor])
  end
end


function Scene:closeShutter(func)
  self.flux:to(self, 1, {shutter = 1.0}):oncomplete(
    function()
      if func then func() end
      G.core.effects:toggle_post_effect(G.noiseeffect, false)
    end)
  love.mouse.setVisible(false)
end

function Scene:openShutter(func)
  self.flux:to(self, 1, {shutter = 0.0}):oncomplete(
    function()
      if func then func() end
      love.mouse.setVisible(true)
    end)
  self.timekeeper:after(0.01, function()
    G.core.effects:toggle_post_effect(G.noiseeffect, true)
  end)
end


function Scene:draw(data)
  love.graphics.push()
  love.graphics.setLineStyle('rough')
  love.graphics.translate(0,30)
  for i,l in ipairs(self.layers) do
    if not self.layerDisabled[i] then
      if i == self.playerLayer then
        self.entities:draw(0)
      else
        love.graphics.draw(l)
      end

      if self.images[i] then
        local t = {}
        for n, i in pairs(self.images[i]) do
          t[#t+1] = i
        end
        table.sort(t, function(a,b) return a[3] < b[3] end)
        for i, k in ipairs(t) do
          love.graphics.draw(k[1], k[2], k[3])
        end
      end
    end
  end

  love.graphics.pop()

  love.graphics.setColor(0,0,0,255)
  love.graphics.rectangle("fill",0,130,341,70)

  love.graphics.setFont(G.font)
  if self.activeDialog and self.activeDialog:isOpen() then
    self.activeDialog:draw()
  else
    local z = self.zone
    if z then
      love.graphics.printf(z[6], 0, 132, 341, "center")
    end
  end
  love.graphics.setColor(0,0,0)
  love.graphics.rectangle("fill", 0, 0, 341, 30)
  love.graphics.setColor(128,128,128)
  love.graphics.setFont(G.titlefont)
  love.graphics.printf(self.title, 2, 5, 341, "center")
  love.graphics.setFont(G.font)

  if self.shutter > 0 then
    love.graphics.stencil(function()
      love.graphics.setColor(255,255,255)
      love.graphics.circle('fill', 170, 100, (1-self.shutter) * 200)
    end, "replace", "1")
    love.graphics.setStencilTest("equal", 0)
    love.graphics.setColor(0,0,0)
    love.graphics.rectangle('fill', 0,0,341, 200)
    love.graphics.setStencilTest()
  end

  if self.textover and self.textover ~= {} then
    local y0 = 100 - 10 * (#self.textover / 2)
    local step = 1 / #self.textover
    for i, l in ipairs(self.textover) do
      local c
      if self.overPos < step * (i/self.scale2) then
        c = 0
      elseif self.overPos > step * (i/self.scale2+1) then
        c = 255
      else
        c = 255 * (self.overPos - step * i/self.scale2) / step 
      end

      love.graphics.setColor(c,c,c)
      love.graphics.printf(l, 0, y0 + (i-1) * 10, 341, "center")
    end
  end

end


function Scene:overlayText(text, scale, scale2)
  self.textover = text
  self.overPos = 0
  self.overFlux = self.flux:to(self, (scale or 1) * #text / Dialog.speechspeed, {overPos=1})
  self.scale2 = scale2 or 5
  self.haveOver = true

end


function Scene:disableLayer(n, yeah)
  self.layerDisabled[n] = yeah
end




return Scene
